/*-
 * Copyright Sergey Kosyakov ks@itp.ac.ru 1999
 * Copyright (C)2002..2020 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#)Copyright Sergey Kosyakov ks@itp.ac.ru 1999"
#ident "@(#)Copyright (C)2002..2020 @BABOLO http://www.babolo.ru/"
#ident "$Id: tund.c,v 1.26 2020/12/05 17:03:19 babolo Exp $"

#define BLIN_COMPAT  3
#define MIFE_COMPAT  5
#define Bpars_COMPAT 3

#include <sys/socket.h>
#include <sys/types.h>
#include <sys/event.h>
#include <sysexits.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sysexits.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <stdio.h>
#include <poll.h>
#include <time.h>
#include <err.h>
#include <netinet/in.h>
#include <openssl/rsa.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <mife.h>
#include "tund.h"
#include "cowords.h"

tund_conf   *tund_cnf;

#if 0
static void
cnfdump(tund_conf *cnf) {
    for (int i = 0; TUND_TABLE_SIZE > i; ++i) {
        if  (0 < cnf->tunTable[i]) printf("%d %u\n", i, cnf->tunTable[i]);
        fflush(stdout);
    }
    printf( "0..%u < %u\n", cnf->stun->length, cnf->stun->maxlen);
    fflush(stdout);
    for (int i = 0; i < cnf->stun->length; ++i) {
        tund_tunnel *tunnel;

        tunnel = cnf->stun->buf[i];
        printf( "#%d* dp=%d socketDivert=%d timer=%d state=%04X tunum=%u flags=%08X\n"
                "f_exp=%"BLIN_X" r_key=%"BLIN_X" localCipherKey=%"BLIN_X" remoteCipherKey=%"BLIN_X"\n"
                "remoteName=%s l_addr=[%u]%u %u.%u.%u.%u:%u to=[%u]%u %u.%u.%u.%u:%u\n"
                "local_key= "
              , i
              , tunnel->dp
              , tunnel->socketDivert
              , tunnel->timer
              , tunnel->state
              , tunnel->tunum
              , tunnel->flags
              , BLIN_I(tunnel->f_exp)
              , BLIN_I(tunnel->r_key)
              , BLIN_I(tunnel->localCipherKey)
              , BLIN_I(tunnel->remoteCipherKey)
              , tunnel->remoteName
              , tunnel->l_addr.sin_len
              , tunnel->l_addr.sin_family
              , (tunnel->l_addr.sin_addr.s_addr >> 24) & 0x00FF
              , (tunnel->l_addr.sin_addr.s_addr >> 16) & 0x00FF
              , (tunnel->l_addr.sin_addr.s_addr >> 8) & 0x00FF
              , tunnel->l_addr.sin_addr.s_addr & 0x00FF
              , tunnel->l_addr.sin_port
              , tunnel->to.sin_len
              , tunnel->to.sin_family
              , (tunnel->to.sin_addr.s_addr >> 24) & 0x00FF
              , (tunnel->to.sin_addr.s_addr >> 16) & 0x00FF
              , (tunnel->to.sin_addr.s_addr >> 8) & 0x00FF
              , tunnel->to.sin_addr.s_addr & 0x00FF
              , ntohs(tunnel->to.sin_port)
              );
        fflush(stdout);
        for (int i = 0; i < 16; ++i) printf(" %02X", tunnel->local_key[i]);
        fflush(stdout);
        printf("\nremote_key= ");
        fflush(stdout);
        for (int i = 0; i < 16; ++i) printf(" %02X", tunnel->remote_key[i]);
        fflush(stdout);
        printf("\n");
        fflush(stdout);
}   }
#endif

static void
usage(int ex) {
    fprintf( stderr
           , "tund @BABOLO V.M "VERS"  "DATE" @BABOLO http://www.babolo.ru/\n"
             "  Using:\n"
             "     tund [-dqv[v...]] [-b HOST] [-c DIR] [-f PATH] [-k PATH] [-p PORT] [-s PATH] [-u PATH]\n"
             "          -b HOST  - bind to IP address of HOST\n"
             "          -c /DIR  - configuration directory\n"
             "          -d       - debug foreground\n"
             "          -E       - give arg of next flag from environment variable\n"
             "          -f PATH  - configuration file\n"
             "          -k PATH  - RSA private key\n"
             "          -p PORT  - IP port name or number\n"
             "          -q       - quiet\n"
             "          -s PATH  - seed\n"
             "          -u PATH  - RSA public key database\n"
             "          -v[v...] - be more verbose\n"
             "          -z TEXT  - recursive flags without variable substitution\n"
             "          -Z TEXT  - recursive flags with variable substitution\n"
             "          -?       - short help\n"
           );
    exit(ex);
}

static int
dobos(tund_conf *cnf, babolo_opts *bos) {
#   define blin_internal_flags (cnf->flags)
    int ex  = EX_OK;

    for (; ;) {
        const char *o;
        int         i;

        i = babolo_getopts(bos);
        if  (!i) break;
        switch (i) {
        case 'b': if  (!(o = babolo_getoptsarg(bos))) usage(EX_USAGE);
                ; if  (!!cnf->bindaddr) {
                ;     ifBLIN_QX1("Redefined -%c option: %s->%s", bos->c, cnf->bindaddr, o);
                ; }
                ; if  (!(cnf->bindaddr = strdup(o))) {
                ;     ifBLIN_QW0("No mem for bindaddr")
                ;     ex = EX_OSERR;
                ;     goto out;
                ; }
                ; break;
        case 'c': if  (!(o = babolo_getoptsarg(bos))) usage(EX_USAGE);
                ; if  ('/' != o[0]) {
                ;     ifBLIN_QX1("Abs path must be used, confdir=%s~ ignored", o);
                ;     errno = EFTYPE;
                ;     continue;
                ; }
                ; if  (!!cnf->confdir) {
                ;     ifBLIN_QX1("Redefined -%c option: %s->%s", bos->c, cnf->confdir, o);
                ; }
                ; if  (!(cnf->confdir = strdup(o))) {
                ;     ifBLIN_QW0("No mem for confdir")
                ;     ex = EX_OSERR;
                ;     goto out;
                ; }
                ; break;
        case 'd': cnf->flags |= TUND_FOREG;
                ; break;
        case 'f': if  (!(o = babolo_getoptsarg(bos))) usage(EX_USAGE);
                ; if  (!!cnf->conf) ifBLIN_QX1("Redefined -%c option: %s->%s", bos->c, cnf->conf, o);
                ; if  (!(cnf->conf = strdup(o))) {
                ;     ifBLIN_QW0("No mem for conf")
                ;     ex = EX_OSERR;
                ;     goto out;
                ; }
                ; break;
        case 'k': if  (!(o = babolo_getoptsarg(bos))) usage(EX_USAGE);
                ; if  (!!cnf->rsak) ifBLIN_QX1("Redefined -%c option: %s->%s", bos->c, cnf->rsak, o);
                ; if  (!(cnf->rsak = strdup(o))) {
                ;     ifBLIN_QW0("No mem for rsak")
                ;     ex = EX_OSERR;
                ;     goto out;
                ; }
                ; break;
        case 'p': if  (!(o = babolo_getoptsarg(bos))) usage(EX_USAGE);
                ; if  (!!cnf->port) ifBLIN_QX1("Redefined -%c option: %s->%s", bos->c, cnf->port, o);
                ; if  (!(cnf->port = strdup(o))) {
                ;     ifBLIN_QW0("No mem for port")
                ;     ex = EX_OSERR;
                ;     goto out;
                ; }
                ; break;
        case 'q': BLIN_QUIET(cnf->flags);
                ; break;
        case 's': if  (!(o = babolo_getoptsarg(bos))) usage(EX_USAGE);
                ; if  (!!cnf->seed) ifBLIN_QX1("Redefined -%c option: %s->%s", bos->c, cnf->seed, o);
                ; if  (!(cnf->seed = strdup(o))) {
                ;     ifBLIN_QW0("No mem for seed")
                ;     ex = EX_OSERR;
                ;     goto out;
                ; }
                ; break;
        case 'u': if  (!(o = babolo_getoptsarg(bos))) usage(EX_USAGE);
                ; if  (!!cnf->rsap) ifBLIN_QX1("Redefined -%c option: %s->%s", bos->c, cnf->rsap, o);
                ; if  (!(cnf->rsap = strdup(o))) {
                ;     ifBLIN_QW0("No mem for rsap")
                ;     ex = EX_OSERR;
                ;     goto out;
                ; }
                ; break;
        case 'v': BLIN_VERBOSE(cnf->flags);
                ; break;
        case '=': blin_ctl(BLIN_CTL_FEQU, cnf->flags);
                  switch((o = babolo_getoptsarg(bos))[0]) {
                  case '0': blin_ctl(BLIN_CTL_DUMP);      exit(EX_OK);
                  case '2': babolo_optext(bos);           exit(EX_OK);
                  case '3': babolo_dumpopts(bos);         exit(EX_OK);
                  }
                ; break;
        case '?': usage(EX_OK);
        default : ifBLIN_QX0("Illegal flag -%c", bos->c);
                ; usage(EX_USAGE);
    }   }
out:
    return(ex);
#   undef blin_internal_flags
}

int
main(int argc, char **argv) {
    tund_conf   *cnf;
    babolo_opts *bos = NULL;
    int          ex  = EX_OK;

    if  (!(cnf = calloc(1, sizeof(tund_conf)))) {
        ifBLIN_QW0("no mem for tund_conf");
        ex = EX_OSERR;
        goto out;
    }
    blin_ctl(BLIN_CTL_FLAG | BLIN_CTL_FSET, (cnf->flags = BLIN_BIT0));
#   define blin_internal_flags (cnf->flags)
    blin_ctl(BLIN_CTL_WHAT | BLIN_CTL_FEQU | BLIN_MASK, BLIN_MODEOAU | BLIN_MODWARN);
    tund_cnf = cnf;
    cnf->pid_fd = -1;
    cnf->port = "9001";
    cnf->pid_nm = TUND_LOCKFILE;
    cnf->confdir = TUND_CONFDIR;
    if  (!(bos = babolo_openopts(Bpars_ESYM | Bpars_SYME, 'Z'))) {
        ifBLIN_QW0("babolo_openopts");
        ex = EX_SOFTWARE;
        goto out;
    }
    if  (!!babolo_setopts(bos, Bpars_NONU, argc, argv, "b:c:df:k:p:qs:u:v=:?")) {
        ifBLIN_QW0("babolo_setopts");
        ex = EX_SOFTWARE;
        goto out;
    }
    if  (!!(ex = dobos(cnf, bos))) {
        ifBLIN_QW0("dobos");
        goto out;
    }
    if  (!cnf->conf) {
        if  (!cnf->confdir) cnf->confdir = TUND_CONFDIR;
        if  (0 > asprintf(&cnf->conf, "%s/%s", cnf->confdir, TUND_CONFILE)) {
            ifBLIN_QW0("no mem for asprintf conf");
            ex = EX_OSERR;
            goto out;
        }
    } else if ('/' != cnf->conf[0]) {
        if  (!cnf->confdir) cnf->confdir = TUND_CONFDIR;
        if  (0 > asprintf(&cnf->conf, "%s/%s", cnf->confdir, cnf->conf)) {
            ifBLIN_QW0("no mem for asprintf conf");
            ex = EX_OSERR;
            goto out;
    }   }
    if  (!(cnf->stun = tund_seq_new(cnf))) {
        ifBLIN_QW0("tund_seq_new");
        ex = EX_OSERR;
        goto out;
    }
    if  (!cnf->confdir) cnf->confdir = TUND_CONFDIR;
    {   off_t            filepos;
        char            *window;
        ssize_t          widew;
        size_t           line;
        babolo_parm     *bp;
        mife_descriptor *md;
        char            *c;

        if  (!(md = mife_init(MIFE_FULL))) {
            ifBLIN_QW0("mife_init");
            ex = EX_IOERR;
            goto out;
        }
        if  (0 > mife_ctlfile(md, cnf->conf)) {
            ifBLIN_QW0("mife_ctlfile %s", cnf->conf);
            ex = EX_IOERR;
            goto out;
        }
        for (filepos = 0, line = 0; ;) {
            if  (!(window = mife_window(md, filepos, 0))) {
                ifBLIN_QW0("mife_window %llu", filepos);
                ex = EX_IOERR;
                goto out;
            }
            if  (0 > (widew = mife_read(md, 0, filepos))) {
                ifBLIN_QW0("mife_read %llu", filepos);
                ex = EX_IOERR;
                goto out;
            }
            if  (5 > widew) break;
            c = window;
            if  (!(bp = babolo_getparm(Bpars_XXLN | Bpars_NOEN | Bpars_NOAL, &c, "#", 0))) {
                ifBLIN_QW0("babolo_getparm in line %lu", line);
                ex = EX_SOFTWARE;
                goto out;
            } else if ((c - window) > widew) {
                ifBLIN_QX0("Owerread for %lu bytes", c - window - widew);
                ex = EX_SOFTWARE;
                errno = EMSGSIZE;
                goto out;
            }
            ifBLIN_QX5("[%d],%s,%s", bp->argc, bp->argv[0], bp->argv[1]);
            if  (bp->argc > 1) {
            ;   if  (!strcmp(bp->argv[0], "tund")) {
            ;       if  (!babolo_setparm(bos, Bpars_NONU, bp, NULL)) {
            ;           ifBLIN_QW0("babolo_setparm in line %lu", line);
            ;           ex = EX_SOFTWARE;
            ;           goto out;
            ;       }
            ;       if  (!!(ex = dobos(cnf, bos))) {
            ;           ifBLIN_QW0("dobos in line %lu", line);
            ;           goto out;
            ;       }
            ;   } else if (!strcmp(bp->argv[0], "tunnel")) {
                    tund_tunnel *tunnel;
                    const char  *word;

            ;       if  (!(tunnel = calloc(1, sizeof(tund_tunnel)))) {
            ;           ifBLIN_QW0("no mem for tunnel");
            ;           ex = EX_OSERR;
            ;           goto out;
            ;       }
            ;       errno = 0;
            ;       tunnel->tunum = strtoul(bp->argv[1], NULL, 0);
            ;       if  (!!errno) {
            ;           ifBLIN_QW0("strtoul in line %lu", line);
            ;           ex = EX_USAGE;
            ;           goto out;
            ;       } else if (65534 < tunnel->tunum) {
            ;           ifBLIN_QX0("Too big tunnel num %u in line %lu", tunnel->tunum, line);
            ;           errno = EDOM;
            ;           ex = EX_USAGE;
            ;           goto out;
            ;       }
            ;       ifBLIN_QX5("tunnel %u", tunnel->tunum);
            ;       for (int i = 2; i < bp->argc; ++i) {
            ;           word = bp->argv[i];
            ;           errno = 0;
            ;           switch(babolo_goword(&cowords, &word)) {
                        case TUND_CF_RULE:
            ;               tunnel->l_addr.sin_port = strtoul(word, NULL, 0);
            ;               if  (!!errno) {
            ;                   ifBLIN_QW0("strtoul in line %lu", line);
            ;                   ex = EX_USAGE;
            ;                   goto out;
            ;               } else if (65534 < tunnel->l_addr.sin_port) {
            ;                   ifBLIN_QX0( "Too big tunnel tag %u in line %lu"
                                          , tunnel->l_addr.sin_port
                                          , line
                                          );
            ;                   errno = EDOM;
            ;                   ex = EX_USAGE;
            ;                   goto out;
            ;               }
            ;               ifBLIN_QX5("tunnel %u tag %u", tunnel->tunum, tunnel->l_addr.sin_port);
            ;               break;
                        case TUND_CF_DVRT:
            ;               tunnel->dp = strtoul(word, NULL, 0);
            ;               if  (!!errno) {
            ;                   ifBLIN_QW0("strtoul in line %lu", line);
            ;                   ex = EX_USAGE;
            ;                   goto out;
            ;               } else if (65534 < tunnel->dp) {
            ;                   ifBLIN_QX0( "Too big tunnel divert port %u in line %lu"
                                          , tunnel->dp
                                          , line
                                          );
            ;                   errno = EDOM;
            ;                   ex = EX_USAGE;
            ;                   goto out;
            ;               }
            ;               ifBLIN_QX5("tunnel %u dvrt %u", tunnel->tunum, tunnel->dp);
            ;               break;
                        case TUND_CF_HOST: {
                                struct addrinfo *hints;
                                struct addrinfo *r;
        int              e;

            ;                   if  (!(hints = calloc(1, sizeof(struct addrinfo)))) {
            ;                       ifBLIN_QW0("no mem for hints");
            ;                       ex = EX_OSERR;
            ;                       goto oul;
            ;                   }
            ;                   hints->ai_family = PF_INET;
            ;                   hints->ai_socktype = SOCK_DGRAM;
            ;                   hints->ai_protocol = IPPROTO_UDP;
            ;                   hints->ai_flags = AI_ADDRCONFIG;
            ;                   if  (!cnf->port) cnf->port = "9001";
            ;                   if  (!!(e = getaddrinfo(word, cnf->port, hints, &r))) {
            ;                       ifBLIN_QW0( "getaddrinfo %s:%s in line %lu: %s"
                                              , word
                                              , cnf->port ? cnf->port : ""
                                              , line
                                              , gai_strerror(e)
                                              );
            ;                       if  (r) freeaddrinfo(r);
            ;                       r = NULL;
            ;                       ex = EX_NOHOST;
            ;                       goto out;
            ;                   }
            ;                   ifBLIN_QX5("l %u f %u", r->ai_addr->sa_len, r->ai_addr->sa_family);
            ;                   bcopy(r->ai_addr, &(tunnel->to), sizeof(struct sockaddr_in));
            ;                   tunnel->remoteName = strdup(word);
            ;               }
            ;               ifBLIN_QX5( "tunnel %u remoteName %s [%u]%u %u.%u.%u.%u:%u?%u"
                                      , tunnel->tunum
                                      , tunnel->remoteName
                                      , tunnel->to.sin_len
                                      , tunnel->to.sin_family
                                      , (tunnel->to.sin_addr.s_addr >> 24) & 0x00FF
                                      , (tunnel->to.sin_addr.s_addr >> 16) & 0x00FF
                                      , (tunnel->to.sin_addr.s_addr >> 8) & 0x00FF
                                      , tunnel->to.sin_addr.s_addr & 0x00FF
                                      , ntohs(tunnel->to.sin_port)
                                      , tunnel->to.sin_port
                                      );
            ;               break;
                        case TUND_CF_ENCR:
            ;               switch(babolo_testword(&cowords, word)) {
                            case TUND_CF_BFIS: tunnel->flags |= TUND_CIPHER_BF;   break;
                            case TUND_CF_IDEA: tunnel->flags |= TUND_CIPHER_IDEA; break;
                            case TUND_CF_RC5:  tunnel->flags |= TUND_CIPHER_RC5;  break;
                            default:
            ;                   ifBLIN_QX0( "Unknown cipher=%s~ for tunnel=%u in line %lu"
                                          , word
                                          , tunnel->tunum
                                          );
            ;                   ex = EX_UNAVAILABLE;
            ;                   errno = ENOPROTOOPT;
            ;                   goto out;
            ;               }
            ;               ifBLIN_QX5("tunnel %u flags 0x%X", tunnel->tunum, tunnel->flags);
            ;               break;
                        default:
            ;               ifBLIN_QX0( "Unknown option=%s~ for tunnel=%u in line %lu"
                                      , bp->argv[i]
                                      , tunnel->tunum
                                      , line
                                      );
            ;               ex = EX_USAGE;
            ;               errno = ENOTSUP;
            ;       }   }
            ;       cnf->tunTable[tunnel->tunum] = 1 + tund_seq_append(cnf, cnf->stun, tunnel);
            ;       ifBLIN_QX3( "tunnel=%u: tag=%d dp=%d host=%s~ flags=0x%X"
                              , tunnel->tunum
                              , tunnel->l_addr.sin_port
                              , tunnel->dp
                              , tunnel->remoteName
                              , tunnel->flags
                              );
            ;   } else {
            ;       ifBLIN_QX1("Ill key %s in line %lu", bp->argv[0], line);
            }   }
            filepos += c - window;
            line += bp->cntr;
            babolo_freeparm(bp);
        }
        if  (!!mife_fini(md)) ifBLIN_QX1("mife_fini");
    }
//  cnfdump(cnf);
    if  (!cnf->port) cnf->port = "9001";
    if  (!cnf->pid_nm) cnf->pid_nm = TUND_LOCKFILE;
    if  (!cnf->rsak && (0 > asprintf(&cnf->rsak, "%s/%s", cnf->confdir, TUND_RSAKEYF))) {
        ifBLIN_QW0("no mem for asprintf rsak");
        ex = EX_OSERR;
        goto out;
    } else if (('/' != cnf->rsak[0]) && (0 > asprintf(&cnf->rsak, "%s/%s", cnf->confdir, cnf->rsak))) {
        ifBLIN_QW0("no mem for asprintf rsak");
        ex = EX_OSERR;
        goto out;
    }
    if  (!cnf->rsap && (0 > asprintf(&cnf->rsap, "%s/%s", cnf->confdir, TUND_RSAPUBDB))) {
        ifBLIN_QW0("no mem for asprintf rsap");
        ex = EX_OSERR;
        goto out;
    } else if (('/' != cnf->rsap[0]) && (0 > asprintf(&cnf->rsap, "%s/%s", cnf->confdir, cnf->rsap))) {
        ifBLIN_QW0("no mem for asprintf rsap");
        ex = EX_OSERR;
        goto out;
    }
    if  (!cnf->seed && (0 > asprintf(&cnf->seed, "%s/%s", cnf->confdir, TUND_SEEDFILE))) {
        ifBLIN_QW0("no mem for asprintf seed");
        ex = EX_OSERR;
        goto out;
    } else if (('/' != cnf->seed[0]) && (0 > asprintf(&cnf->seed, "%s/%s", cnf->confdir, cnf->seed))) {
        ifBLIN_QW0("no mem for asprintf rsak");
        ex = EX_OSERR;
        goto out;
    }
    ifBLIN_QX5("bindaddr=%s~ port=%s~", cnf->bindaddr, cnf->port);
    {   struct addrinfo *hints;
        struct addrinfo *r;
        int              e;

        if  (!(hints = calloc(1, sizeof(struct addrinfo)))) {
            ifBLIN_QW0("no mem for hints");
            ex = EX_OSERR;
            goto out;
        }
        hints->ai_family = PF_INET;
        hints->ai_socktype = SOCK_DGRAM;
        hints->ai_protocol = IPPROTO_UDP;
        hints->ai_flags = AI_ADDRCONFIG | AI_PASSIVE;
        if  (!!(e = getaddrinfo(cnf->bindaddr, cnf->port, hints, &r))) {
            ifBLIN_QW0( "getaddrinfo %s:%s: %s"
                      , cnf->bindaddr ? cnf->bindaddr : ""
                      , cnf->port ? cnf->port : ""
                      , gai_strerror(e)
                      );
            if  (r) freeaddrinfo(r);
            r = NULL;
            ex = EX_NOHOST;
            goto out;
        }
        if  (0 > (cnf->isock = socket(r->ai_family, r->ai_socktype, r->ai_protocol))) {
            /* There is list by r->ai_next pointer XXXX */
            ifBLIN_QW0( "IN socket %s:%s"
                      , cnf->bindaddr ? cnf->bindaddr : ""
                      , cnf->port ? cnf->port : ""
                      );
            ex = EX_NOHOST;
            goto out;
        }
        if  (0 > bind(cnf->isock, r->ai_addr, r->ai_addrlen)) {
            ifBLIN_QW0( "IN bind %s:%s"
                      , cnf->bindaddr ? cnf->bindaddr : ""
                      , cnf->port ? cnf->port : ""
                      );
            ex = EX_NOHOST;
            goto out;
        }
        if  (!cnf->s_port) cnf->s_port = ((struct sockaddr_in *)r->ai_addr)->sin_port;
        cnf->s_addr = ((struct sockaddr_in *)r->ai_addr)->sin_addr.s_addr;
        if  (r) freeaddrinfo(r);
        r = NULL;
        free(hints);
    }
    if  (fcntl(cnf->isock, F_SETFL, FNDELAY) == -1) {                                /* Non-blocking */
        ifBLIN_QW0( "IN fcntl FNDELAY %s:%s"
                  , cnf->bindaddr ? cnf->bindaddr : ""
                  , cnf->port ? cnf->port : ""
                  );
        ex = EX_NOHOST;
        goto out;
    }
    ifBLIN_QX2( "IN socket %d = %u.%u.%u.%u:%u"
              , cnf->isock
              , cnf->s_addr & 0xFF
              , (cnf->s_addr >> 8) & 0xFF
              , (cnf->s_addr >> 16) & 0xFF
              , (cnf->s_addr >> 24) & 0xFF
              , ntohs(cnf->s_port)
              );
    if  (!cnf->s_port || (0 >= cnf->stun->length)) {
        ifBLIN_QX0("%u %u", cnf->s_port, cnf->stun->length);
        errno = EINVAL;
        ex = EX_USAGE;
        goto out;
    }
    if  (!!(ex = tund_init_rng(cnf))) {
        ifBLIN_QW0("tund_init_rng=%s~", cnf->seed);
        goto out;
    }
    if  (!!(ex = tund_init_secur(cnf))) {
        ifBLIN_QW0("tund_init_secur=%s~", cnf->rsak);
        goto out;
    }
    if  (!!(ex = tund_init_alarm(cnf))) {
        ifBLIN_QW0("tund_init_alarm");
        goto out;
    }
    if  (!(cnf->flags & TUND_FOREG)) {
        if  (daemon(0, 0)) {
            ifBLIN_QW0("daemon");
            ex = EX_OSERR;
            goto out;
        }
        blin_ctl(BLIN_CTL_WHAT | BLIN_CTL_FEQU | BLIN_MASK, BLIN_MODLOGR);
    }
    if  (!!unlink(cnf->pid_nm)) ifBLIN_QW1("unlink PID %s", cnf->pid_nm);
    if  (0 > (cnf->pid_fd = open(cnf->pid_nm, O_CREAT | O_TRUNC | O_WRONLY | O_EXLOCK, 0666))) {
        ifBLIN_QW0("open PID %s", cnf->pid_nm);
        ex = EX_IOERR;
        goto oul;
    }
    {   char *pid;
        int i;

        i = asprintf( &pid, "%d\n%u.%u.%u.%u:%u\n"
                    , getpid()
                    , cnf->s_addr & 0xFF
                    , (cnf->s_addr >> 8) & 0xFF
                    , (cnf->s_addr >> 16) & 0xFF
                    , (cnf->s_addr >> 24) & 0xFF
                    , ntohs(cnf->s_port)
                    );
        if  (!pid || (i < 0)) {
            ifBLIN_QW0("asprintf");
            ex = EX_IOERR;
            goto oul;
        }
        if  (mife_writ(cnf->pid_fd, pid, 0) != i) {
            ifBLIN_QW0("mife_writ %s", cnf->pid_nm);
            ex = EX_IOERR;
            goto oul;
        }
        free(pid);
    }
    {   struct sockaddr_in saddr;
        tund_tunnel       *tun;
        int                sd;
        tund_tunnel       *t;

        for (int tun_ind = 0; tun_ind < cnf->stun->length; tun_ind++) {
            /* Set up tunnel(s) divert sockets */
            tun = (tund_tunnel *)cnf->stun->buf[tun_ind];
            for (int i = 0; i < tun_ind; i++) {
                t = (tund_tunnel *)cnf->stun->buf[i];
                if  (t->dp == tun->dp) {
                    ifBLIN_QX0("Illegal configuration: tunnels should have different divert sockets");
                    ex = EX_USAGE;
                    goto oul;
                }
                if  (t->tunum == tun->tunum) {
                    ifBLIN_QX0("Illegal configuration: tunnels should have different labels");
                    ex = EX_USAGE;
                    goto oul;
            }   }
            bzero(&saddr, sizeof(saddr));
            saddr.sin_port = htons((u_int16_t)tun->dp);
            saddr.sin_family = AF_INET;
            if  (0 > (sd = socket(PF_INET, SOCK_RAW, IPPROTO_DIVERT))) {
                ifBLIN_QW0("socket PF_INET SOCK_RAW IPPROTO_DIVERT");
                ex = EX_IOERR;
                goto oul;
            }
            if  (bind(sd, (struct sockaddr *)&saddr, sizeof(saddr))) {
                ifBLIN_QW0("bind %d to divert port %d", sd, tun->dp);
                ex = EX_IOERR;
                goto oul;
            }
            if  (0 > (fcntl(sd, F_SETFL, FNDELAY))) {                         /* Set to non-blocking */
                ifBLIN_QW0("fcntl %d F_SETFL FNDELAY on divert port %d", sd, tun->dp);
                ex = EX_IOERR;
                goto oul;
            }
            tun->socketDivert = sd;
            ifBLIN_QX4( "Divert port %d is binded to socket %d for t=%d"
                      , tun->dp, tun->socketDivert, tun_ind
                      );
    }   }
    {   struct pollfd *fds;
        tund_tunnel   *tun;
        int            nfd;
        int            ret;
        tund_CBF      *cb;

        nfd = cnf->stun->length + 1;
        if  (!(fds = calloc(nfd, sizeof(struct pollfd)))) {
            ifBLIN_QW0("no memory for pollfd");
            ex = EX_OSERR;
            goto oul;
        }
        if  (!(cb = calloc(nfd, sizeof(tund_CBF)))) {
            ifBLIN_QW0("no memory for callbacks");
            ex = EX_OSERR;
            goto oul;
        }
        for (int i = 0; i < cnf->stun->length; i++) {
            tun = (tund_tunnel *)cnf->stun->buf[i];
            fds[i].fd = tun->socketDivert;
            fds[i].events = POLLRDNORM;
            cb[i].tun_ind = i;
            cb[i].f = tund_inp_dvrt;
            if  (!(tun->flags & TUND_CIPHER_MASK)) {
                tun->state = TUND_S_UP;
            } else if (tund_start_sectun(cnf, i)) {          /* Initiate encryption for that tunnel */
                ifBLIN_QW1("No start tunel %d", i);
        }   }
        fds[cnf->stun->length].fd = cnf->isock;
        fds[cnf->stun->length].events = POLLRDNORM;
        cb[cnf->stun->length].f = tund_read_pkt;
        cb[cnf->stun->length].tun_ind = -1;                                /* Does have no any means */
        for (; ;) {                                                                     /* Main loop */
            if  (0 > (ret = poll(fds, nfd, 15000))) {
                if  (errno == EINTR) continue;
                ifBLIN_QW0("poll in main loop");
                ex = EX_IOERR;
                goto oul;
            }
            ifBLIN_QX4("poll in main loop %d", ret);
            for (int i = 0; i < nfd; i++) {
                if  (fds[i].revents) {
                    ifBLIN_QX4("Event 0x%x on fd %d", fds[i].revents, fds[i].fd);
                    if  (!!cb[i].f(cnf, cb[i].tun_ind, fds[i].fd)) {
                        ifBLIN_QX0("callback");
    }   }   }   }   }
oul:
    flock(cnf->pid_fd, LOCK_UN);
    close(cnf->pid_fd);
    cnf->pid_fd = -1;
out:
    exit(ex);
#   undef blin_internal_flags
}
